"""
* GitHub Utilities
* Generalized utilities for accessing GitHub files or GitHub API endpoints.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
import os
import zipfile
from logging import getLogger
from pathlib import Path
import requests
from typing import Optional, Union, Callable

# Third Party Imports
import yarl
from backoff import on_exception, expo
from ratelimit import RateLimitDecorator, sleep_and_retry

# Local Imports
from omnitils.fetch import download_file
from omnitils.files.folders import mkdir_full_perms
from omnitils.fetch._core import request_header_default, chunk_size_default

# Rate limiter to safely limit GitHub requests
github_rate_limit = RateLimitDecorator(calls=60, period=3600)
github_rate_limit_authenticated = RateLimitDecorator(calls=7, period=5)

"""
* Handlers
"""


def gh_request_handler(func) -> Callable:
    """Wrapper for a GitHub request function to handle retries and rate limits on
    unauthenticated requests (60 per hour)."""
    @sleep_and_retry
    @github_rate_limit
    @on_exception(expo, requests.exceptions.RequestException, max_tries=2, max_time=1)
    def decorator(*args, **kwargs):
        return func(*args, **kwargs)
    return decorator


def gh_request_handler_authenticated(func) -> Callable:
    """Wrapper for a GitHub request function to handle retries and rate limits on
    authenticated requests (5000 per hour)."""
    @sleep_and_retry
    @github_rate_limit_authenticated
    @on_exception(expo, requests.exceptions.RequestException, max_tries=2, max_time=1)
    def decorator(*args, **kwargs):
        return func(*args, **kwargs)
    return decorator


"""
* Request Utilities
"""


def gh_get_header(
    header: dict | None = None,
    auth_token: str | None = None
):
    """Uses the provided header or falls back to default, then injects auth token if valid.

    Args:
        header: Header object to pass with request, uses default if not provided.
        auth_token: Auth token to inject into header if provided.

    Returns:
        Header to pass with a request.
    """
    header = header or request_header_default.copy()
    if auth_token:
        header['Authorization'] = f'token {auth_token}'
    return header


def gh_get_data_json(
    url: Union[str, yarl.URL],
    header: dict | None = None,
    auth_token: str | None = None,
    handler: Optional[Callable] = None
) -> dict | list | tuple:
    """Request a manifest file and return its JSON loaded data.

    Args:
        url: URL to the data file resource hosted on GitHub.
        header: Header object to pass with request, uses default if not provided.
        auth_token: GitHub auth token to inject into header if provided, raises rate limits.
        handler: Decorator function to handle retries, rate limits, etc. Uses built-in `request_handler_github`
            decorator if not provided.

    Raises:
        RequestException if resource could not be returned.
    """
    # Use provided header or choose one
    handler = handler or (
        gh_request_handler if auth_token
        else gh_request_handler_authenticated)

    @handler
    def _make_request(_url, _header, _token):
        with requests.get(_url, headers=gh_get_header(_header, _token)) as r:
            r.raise_for_status()
            return r.json()
    return _make_request(url, header, auth_token)


"""
* Download Utilities
"""


def gh_download_file(
    url: str | yarl.URL,
    path: Path,
    header: dict | None = None,
    auth_token: str | None = None,
    chunk_size: int = chunk_size_default,
    handler: Optional[Callable] = None
):
    """Requests a file and saves it to a given path location.

    Args:
        url: URL to the file hosted on GitHub.
        path: Absolute path to save the file to.
        header: Header object to pass with request, uses default if not provided.
        auth_token: GitHub auth token to inject into header if provided, raises rate limits.
        chunk_size: Size of each chunk to write when using iter_content to save file.
        handler: Decorator function to handle retries, rate limits, etc. Uses built-in `request_handler_github`
            decorator if not provided.

    Returns:
        Path to the file.
    """
    # Use provided header or choose one
    handler = handler or (
        gh_request_handler if auth_token
        else gh_request_handler_authenticated)

    @handler
    def _make_request():
        return download_file(
            url=url,
            path=path,
            header=gh_get_header(header, auth_token),
            chunk_size=chunk_size)

    # Download file
    _make_request()
    return path


def gh_download_repository(
    user: str,
    repo: str,
    path: Path,
    branch: str = 'main',
    header: dict | None = None,
    auth_token: str | None = None,
    chunk_size: int = chunk_size_default,
    handler: Optional[Callable] = None
) -> Path:
    """Download a gitHub repository and extract it to the provided path.

    Args:
        user: Username of the repository owner.
        repo: Name of the GitHub repository.
        path: Path to download and extract the repository to.
        branch: Name of the repository branch to download.
        header: Header object to pass with request, uses default if not provided.
        auth_token: GitHub auth token to inject into header if provided, raises rate limits.
        chunk_size: Size of each chunk to write when using iter_content to save file.
        handler: Decorator function to handle retries, rate limits, etc. Uses built-in `request_handler_github`
            decorator if not provided.

    Returns:
        Path to the repository.
    """
    # Establish base values
    temp_file = path / 'temp.zip'

    # Download the archive
    temp_file = gh_download_file(
        url=f"https://github.com/{user}/{repo}/archive/refs/heads/{branch}.zip",
        path=temp_file,
        header=header,
        auth_token=auth_token,
        chunk_size=chunk_size,
        handler=handler)

    # Extract files, remove temporary zip
    with zipfile.ZipFile(temp_file, 'r') as zip_ref:
        zip_ref.extractall(path=path)
    os.remove(temp_file)
    return Path(path, f'{repo}-{branch}')


def gh_download_directory_files(
    user: str,
    repo: str,
    repo_dir: str,
    path: Path,
    file_type: str | None = None,
    header: dict | None = None,
    auth_token: Optional[str] = None,
    chunk_size: int = chunk_size_default,
    handler: Optional[Callable] = None
) -> list[Path]:
    """Download all files from a specific directory in a GitHub repository to a given path.

    Args:
        user: Username of the repository owner.
        repo: Name of the GitHub repository.
        repo_dir: Directory location in the GitHub repository to download files from.
        path: Path to save the files to.
        file_type: File type files must match in order to be downloaded, if provided.
        header: Header object to pass with request, uses default if not provided.
        auth_token: Optional GitHub personal access token to use, for increasing rate limits.
        chunk_size: Size of each chunk to write when using iter_content to save file.
        handler: Decorator function to handle retries, rate limits, etc. Uses built-in `request_handler_github`
            decorator if not provided.

    returns:
        A list containing a Path to each file downloaded.
    """
    # Ensure path exists
    if not path.is_dir():
        mkdir_full_perms(path)
    url = f'https://api.github.com/repos/{user}/{repo}/contents/{repo_dir}'
    header = gh_get_header(header, auth_token)
    files: list[Path] = []

    # Get file data from API
    try:
        data = gh_get_data_json(
            url=url,
            header=header,
            auth_token=auth_token,
            handler=handler)
    except requests.RequestException:
        # Couldn't get GitHub directory info
        getLogger().warning(
            f"Couldn't retrieve GitHub directory information!\n"
            f"URL: {url}")
        return files

    # Download each file
    for file in data:
        # Is this a file?
        if not all([bool(n in file) for n in ['type', 'name', 'download_url']]):
            continue
        # Is the filetype valid?
        if file['type'] != 'file' or (file_type and not file['name'].endswith(file_type)):
            continue

        # Download file
        try:
            files.append(
                gh_download_file(
                    url=file['download_url'],
                    path=Path(path, file['name']),
                    header=header,
                    chunk_size=chunk_size,
                    handler=handler))
        except requests.RequestException:
            # Couldn't download file
            getLogger().warning(
                f"Download failed: {file['download_url']}")
            continue
    return files



"""
* Enums Utilities
* Generalized utility Enum classes.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
from enum import Enum, EnumMeta
from typing import Iterator, Any

# Local Imports
from omnitils.properties import default_prop


"""
* String Enum Classes
"""


class ConstantMeta(EnumMeta):
    """Metaclass for constant enums."""

    def __iter__(cls) -> Iterator[Any]:
        """Iterate over the Enum collection."""
        for n in cls.__members__.values():
            yield n.value

    def __contains__(cls, item: Any) -> bool:
        """Check if item is contained in Enum collection."""
        return bool(item in cls.__members__.values())

    def items(cls) -> Iterator[tuple[str, Any]]:
        """Iterate over the names and values contained in the Enum collection."""
        for k, v in cls.__members__.items():
            yield k, v.value

    @default_prop
    def Default(cls) -> Any:
        """Allow for a 'default' Enum in the collection, overwrite this to manually define a default
            value."""
        return cls._value2member_map_.__iter__().__next__()


class StrConstantMeta(EnumMeta):
    """Metaclass for string constant enums."""

    def __contains__(cls, item: str) -> bool:
        """Check if item is contained in Enum collection."""
        return bool(item in cls.__members__.values())

    def __iter__(cls) -> Iterator[str]:
        """Iterate over the Enum collection."""
        for n in cls.__members__.values():
            yield n.value

    def items(cls) -> Iterator[tuple[str, str]]:
        """Iterate over the names and values contained in the Enum collection."""
        for k, v in cls.__members__.items():
            yield k, v.value

    @default_prop
    def Default(cls) -> str:
        """Allow for a 'default' Enum in the collection, overwrite this to manually define a default
            value."""
        return cls._value2member_map_.__iter__().__next__()


"""
* String Enum Classes
"""


class StrConstant(str, Enum, metaclass=StrConstantMeta):
    """Enum constant where the value is always a string."""

    def __str__(self) -> str:
        """Use cached value for string representation."""
        return self.value

    @default_prop
    def value(self) -> str:
        """Cache each value and cast as a string."""
        return str(self._value_)



"""
* Module Utilities
* General utilities for dynamically loading and managing Python modules.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
import importlib
import os
from importlib.util import spec_from_file_location, module_from_spec
from pathlib import Path
import sys
from types import ModuleType
from typing import Union, Optional

"""
* Types
"""

ModuleTree = dict[str, Union[ModuleType, 'ModuleTree']]

"""
* Sys Utils
"""


def add_python_path(path: Union[str, os.PathLike]) -> None:
    """Add a path to 'sys.path', allowing relative import discovery.

    Args:
        path: Directory path to add to `sys.path`.
    """
    p = str(path)
    if p not in sys.path:
        sys.path.append(p)


def remove_python_path(path: Union[str, os.PathLike]) -> None:
    """Remove a path from 'sys.path' to keep the namespace clean.

    Args:
        path: Directory path to remove from `sys.path`.
    """
    p = str(path)
    if p in sys.path:
        sys.path.remove(p)


"""
* Packages
"""


def import_package(name: str, path: Path, hotswap: bool = False) -> Optional[ModuleType]:
    """Loads a module package using importlib or from 'sys.modules', allowing for forced reloads.

    Args:
        name: Name of the module.
        path: Path to the module.
        hotswap: Whether to force the module to be reloaded fresh.

    Returns:
        The loaded module.
    """
    # Return previously loaded module
    if name in sys.modules:
        if not hotswap:
            return sys.modules[name]
        del sys.modules[name]

    # Add parent folder to path
    add_python_path(str(path))

    # Collect nested modules before importing and executing
    import_nested_modules(
        name=name,
        path=path,
        recursive=True,
        hotswap=hotswap,
        ignored=['__init__.py'])

    # Return None if init module not found
    init_module = path / '__init__.py'
    if not init_module.is_file():
        remove_python_path(str(path))
        return

    # Import and execute the init module
    module = import_module_from_path(name=name, path=init_module, hotswap=hotswap)

    # Reset system paths and return
    remove_python_path(str(path))
    return module


"""
* Dynamic Modules
"""


def import_nested_modules(
    name: str,
    path: Path,
    recursive: bool = True,
    hotswap: bool = False,
    ignored: Optional[list[str]] = None
) -> ModuleTree:
    """Imports modules nested inside a directory.

    Args:
        name: Module name to give the directory.
        path: Path to a directory to import as a module.
        recursive: Whether to recursively import nested directories as submodules.
        hotswap: Whether to force modules to be imported fresh, if any were imported previously.
        ignored: Submodules to ignore when importing.

    Returns:
        A dictionary tree of modules imported.
    """
    ignored = ignored or []
    imported = {name: []}
    for item in os.listdir(path):
        if item in ['__pycache__', *ignored]:
            continue

        # Establish submodule path and name
        p = path / item
        n = '.'.join([name, p.stem])

        # Nested folder or python file?
        if p.is_dir() and recursive:
            # Import directory
            add_python_path(str(p))
            module = import_nested_modules(
                name=n, path=p, hotswap=hotswap, ignored=ignored)
            if module:
                imported[n] = module
            remove_python_path(str(p))
        elif p.is_file() and p.suffix == '.py':
            # Import module
            imported[n] = import_module_from_path(name=n, path=p, hotswap=hotswap)
    return imported


def import_module_from_path(name: str, path: Path, hotswap: bool = False) -> ModuleType:
    """Import a module from a given path.

    Args:
        name: Name of the module.
        path: Path to the module.
        hotswap: Whether to

    Returns:
        Loaded and executed module.
    """
    # Return previously loaded module
    if name in sys.modules:
        if not hotswap:
            return sys.modules[name]
        del sys.modules[name]

    # Import and execute the module
    spec = spec_from_file_location(name=name, location=path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    sys.modules[name] = module
    return module


"""
* Local Modules
"""


def get_local_module(module_path: str, hotswap: bool = False) -> ModuleType:
    """Lookup a local module, import it if no previously imported or hotswap is enabled, then return it.

    Args:
        module_path: Path to the module in module notation, e.g. "src.templates"
        hotswap: If True, always re-import module if previously loaded.

    Returns:
        ModuleType: Loaded module.

    Raises:
        ImportError: If module couldn't be loaded successfully.
    """
    # Check if the module has already been loaded
    if module_path in sys.modules:
        if not hotswap:
            return sys.modules[module_path]
        del sys.modules[module_path]

    # Load the module fresh and cache it
    module = importlib.import_module(module_path)
    sys.modules[module_path] = module
    return module



"""
* Fetch Utilities
* Generalized utilities for fetching live data.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
from omnitils.fetch._core import *
from omnitils.fetch.download import *



"""
* Exception Utilities
* Generalized utilities for dealing with exceptions.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
from contextlib import suppress
from logging import getLogger
from typing import Callable, Any, Optional

"""
* Utility Decorators
"""


def log_on_exception(logr: Any = None) -> Callable:
    """Decorator to log any exception that occurs.

    Args:
        logr: Logger object to output any exception messages.

    Returns:
        Wrapped function.
    """
    logr = logr or getLogger()

    def decorator(func):
        def wrapper(*args, **kwargs):
            # Final exception catch
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logr.log_exception(e)
                raise e
        return wrapper
    return decorator


def return_on_exception(response: Optional[Any] = None) -> Callable:
    """Decorator to handle any exception and return appropriate failure value.

    Args:
        response: Value to return if an exception occurs.

    Returns:
        Wrapped function.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            # Final exception catch
            with suppress(Exception):
                return func(*args, **kwargs)
            return response
        return wrapper
    return decorator



"""
* Schema Utilities
* Generalized utilities for creating and validating schemas.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
from typing import Any, Optional

# Third Party Imports
from pydantic import BaseModel

"""
* Types
"""


PriorityMap = dict[str, list[str | tuple[str, str] | tuple[None, Any] | None]]


"""
* Schema Classes
"""


class Schema(BaseModel):
    """Basic schema class, extending the Pydantic BaseModel."""


class ArbitrarySchema(BaseModel):
    """Schema class allowing for arbitrary types."""

    class Config:
        arbitrary_types_allowed = True


class DictSchema(Schema):
    """Dictionary schema class. Returns all new instances as a dictionary."""

    def __new__(cls, **data):
        """Return new instance as a dictionary."""
        new = super().__new__(cls)
        new.__init__(**data)
        return new.model_dump()



"""
* Dict Utilities
* Generalized utilities for working with dictionaries.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
from typing import Hashable, Any

"""
* Dict Inversions
"""


def invert_map(d: dict[Hashable, Hashable]) -> dict[Hashable, Hashable]:
    """Flips the key, val in a dictionary to val, key.

    Args:
        d: Dictionary where the values are hashable.

    Returns:
        Reversed dictionary.
    """
    inverted = {}
    for k, v in d.items():
        inverted.setdefault(v, k)
    return inverted


def invert_map_multi(d: dict[Hashable, Hashable]) -> dict[Hashable, list[Hashable]]:
    """Flips the key, val in a dictionary to val, [keys], preserving cases where the same
    value is mapped to multiple keys.

    Args:
        d: Dictionary where the values are hashable.

    Returns:
        Reversed dictionary.
    """
    inverted = {}
    for k, v in d.items():
        inverted.setdefault(v, []).append(k)
    return inverted


"""
* Dict Sorting
"""


def sort_by_key(d: dict, reverse: bool = False) -> dict:
    """Sort a dictionary by its key.

    Args:
        d: Dictionary to sort by its key.
        reverse: Whether to reverse the sorting order.

    Returns:
        Key sorted dictionary.
    """
    return dict(sorted(d.items(), reverse=reverse))


def sort_by_value(d: dict, reverse: bool = False) -> dict:
    """Sort a dictionary by its value.

    Args:
        d: Dictionary to sort by its value.
        reverse: Whether to reverse the sorting order.

    Returns:
        Value sorted dictionary.
    """
    return dict(sorted(d.items(), key=lambda item: item[1], reverse=reverse))


"""
* Dict Data Utils
"""


def next_item(d: dict) -> tuple[Hashable, Any]:
    """Extract the first key, value pair from a dictionary.

    Args:
        d: Dictionary to retrieve first key, value pair from.

    Returns:
        Tuple containing key, value.
    """
    return next(iter(d.items()))


def last_item(d: dict) -> tuple[Hashable, Any]:
    """Extract the last key, value pair from a dictionary.

    Args:
        d: Dictionary to retrieve last key, value pair from.

    Returns:
        Tuple containing key, value.
    """
    return list(d.items())[-1]



"""
* Property Utilities
* General utility property decorators.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
from typing import Callable


"""
* Func Based Properties
"""


def auto_prop(func: Callable) -> property:
    """Property decorator wrapper which automatically creates a setter."""
    attr_type = func.__annotations__.get('return', str) if (
        hasattr(func, '__annotations__')) else str
    auto_name = f"_{func.__name__}"

    def getter(self) -> attr_type:
        """Getter for retrieving the value of the implied attribute."""
        return getattr(self, auto_name)

    def setter(self, value: attr_type) -> None:
        """Setter for changing the value of the implied attribute."""
        setattr(self, auto_name, value)

    # Return complete property
    return property(getter, setter, doc=func.__doc__)


def default_prop(func: Callable) -> property:
    """Alternative to the `@cached_property` decorator which:
        - Uses the wrapped function as a default value for the property.
        - Caches the default value when it is first accessed.
        - Allows the value to be changed later and caches the new value.
        - Allows the value to be deleted, which reroutes it to the default value.
    """
    attr_type = func.__annotations__.get('return', str) if (
        hasattr(func, '__annotations__')) else str
    cache_name = f"_{func.__name__}"

    def getter(self) -> attr_type:
        """Wrapper for getting cached value. If value doesn't exist, initialize it."""
        try:
            return getattr(self, cache_name)
        except AttributeError:
            value = func(self)
            setattr(self, cache_name, value)
            return value

    def setter(self, value: attr_type) -> None:
        """Setter for invalidating the property cache and caching a new value."""
        setattr(self, cache_name, value)

    def deleter(self) -> None:
        """Deleter for invalidating the property cache."""
        if hasattr(self, cache_name):
            delattr(self, cache_name)

    # Return complete property
    return property(getter, setter, deleter, getattr(func, '__doc__', ''))


"""
* Class Based Properties
"""


class tracked_prop:
    """A property which can be tracked for changes in the `_changes` attribute of the parent object.

    Notes:
        Requires the parent class or instance to define a class or instance attribute '_changes' as
            an empty set, e.g. `_changes = set()`.
    """

    def __init__(self, getter):
        """Initializes the property.

        Args:
            getter: Decorated method which acts as a getter for the property's default value.
        """
        self.getter = getter
        self._value = None
        self._name = None

    def __get__(self, instance, owner):
        """Either sets and returns the default value or returns the cached value.

        Args:
            instance: Instance of the parent object for the decorated getter.
            owner: Class of the parent object.

        Returns:
            Cached or default value.
        """
        if self._value is None:
            self._value = self.getter(instance)
        return self._value

    def __set__(self, instance, value) -> None:
        """Sets a new cached value for the decorated property and adds it to changes.

        Args:
            instance: Instance of the parent object.
            value: Value to set for the cached property.
        """
        self._value = value
        instance._changes.add(self._name)
        return

    def __call__(self, func):
        """Called when decorator is used aas a function.

        Args:
            func: Function to be decorated.

        Returns:
            An instance of this decorator class.
        """
        self.getter = func
        return self

    def __set_name__(self, owner, name) -> None:
        """Sets the name of the decorated property.

        Args:
            owner: Class of the parent object.
            name: Name of the decorated property.
        """
        self._name = name

    def __delete__(self, owner):
        """Clears the cached value of the decorated property.

        Args:
            owner: Class of the parent object.
        """
        self._value = None



"""
* Utils: Compressing and Decompressing Archives
"""
# Standard Library
import bz2
from contextlib import suppress
import gzip
import gc
import lzma
import os
from pathlib import Path
import shutil
import subprocess
import tarfile
from threading import Lock
from typing import Optional, Callable
import zipfile

# Third Party Imports
from loguru import logger
from tqdm import tqdm
import py7zr
from py7zr import SevenZipFile, FILTER_LZMA, FILTER_X86

# Local Imports
from omnitils.enums import StrConstant
from omnitils.strings import str_to_bool_safe

# Locking mechanism
ARCHIVE_LOCK = Lock()

"""
* Enums
"""


class ArchType(StrConstant):
    """Recognized archive extension types."""
    Zip = '.zip'
    GZip = '.gz'
    XZip = '.xz'
    BZip2 = '.bz2'
    SevenZip = '.7z'
    TarGZip = '.tar.gz'
    TarXZip = '.tar.xz'
    TarBZip2 = '.tar.bz2'
    TarSevenZip = '.tar.xz'


class WordSize(StrConstant):
    """Word Size for 7z compression."""
    WS16 = "16"
    WS24 = "24"
    WS32 = "32"
    WS48 = "48"
    WS64 = "64"
    WS96 = "96"
    WS128 = "128"


class DictionarySize(StrConstant):
    """Dictionary Size for 7z compression."""
    DS32 = "32"
    DS48 = "48"
    DS64 = "64"
    DS96 = "96"
    DS128 = "128"
    DS192 = "192"
    DS256 = "256"
    DS384 = "384"
    DS512 = "512"
    DS768 = "768"
    DS1024 = "1024"
    DS1536 = "1536"


"""
* Compression Utils
"""


def compress_7z_py(
    path_in: Path,
    path_out: Optional[Path] = None,
    filters: Optional[list[dict[str, int]]] = None
) -> Optional[Path]:
    """Compress a target file to a target 7z archive using the py7zr module.

    Args:
        path_in: File to compress.
        path_out: Path to the archive to be saved. Use 'compressed' subdirectory if not provided.
        filters: Filters used when initializing the SevenZipFile object, uses LZMA+BCJ by default.

    Returns:
        Path to the resulting 7z archive if successful, otherwise None.
    """
    lzma_bcj = filters or [{'id': FILTER_X86}, {'id': FILTER_LZMA}]
    with SevenZipFile(path_out, 'w', filters=lzma_bcj) as z:
        z.write(path_in)
    return path_out


def compress_7z_7zip(
    path_in: Path,
    path_out: Optional[Path] = None,
    compress_level: int = 9,
    word_size: WordSize = WordSize.WS16,
    dict_size: DictionarySize = DictionarySize.DS1536,

) -> Optional[Path]:
    """Compress a target file to a target 7z archive using the 7-Zip CLI.

    Notes:
        Compressing using the 7-Zip CLI is relevantly faster than compressing with the py7zr
            package, but required 7-Zip be installed on the host system.

    Args:
        path_in: File to compress.
        path_out: Path to the archive to be saved. Use 'compressed' subdirectory if not provided.
        compress_level: Compression level to use (1 to 9), default is 9.
        word_size: Word size value to use for the compression, default is 16.
        dict_size: Dictionary size value to use for the compression, default is 1536.

    Returns:
        Path to the resulting 7z archive if successful, otherwise None.
    """
    null_device = open(os.devnull, 'w')
    subprocess.run([
        "7z", "a", "-t7z", "-m0=LZMA",
        f"-mx={compress_level}",
        f"-md={dict_size}M",
        f"-mfb={word_size}",
        str(path_out),
        str(path_in)
    ], stdout=null_device, stderr=null_device)
    return path_out


def compress_7z(
    path_in: Path,
    path_out: Optional[Path] = None,
    use_7zip: bool = False,
    compress_level: int = 9,
    word_size: WordSize = WordSize.WS16,
    dict_size: DictionarySize = DictionarySize.DS1536,
) -> Optional[Path]:
    """Compress a target file and save it as a 7z archive to the output directory.

    Args:
        path_in: File to compress.
        path_out: Path to the archive to be saved. Use 'compressed' subdirectory if not provided.
        use_7zip: Whether to use the 7zip CLI to perform the compression, defaults to False. Can also be flagged
            using the USE_7ZIP environment variable (string bool).
        compress_level: Compression level to use (1 to 9). Only used with 7-Zip CLI, default is 9.
        word_size: Word size value to use for the compression. Only used with 7-Zip CLI, default is 16.
        dict_size: Dictionary size value to use for the compression. Only used with 7-Zip CLI, default is 1536.

    Returns:
        Path to the resulting 7z archive if successful, otherwise None.
    """
    # Define the output file path
    path_out = path_out or Path(path_in.parent, '.compressed', path_in.name)
    path_out = path_out.with_suffix('.7z')

    # Compress the file
    with suppress(Exception):

        # Compress with 7-Zip
        if use_7zip or str_to_bool_safe(os.environ.get('USE_7ZIP', '0')):
            return compress_7z_7zip(
                path_in=path_in,
                path_out=path_out,
                compress_level=compress_level,
                word_size=word_size,
                dict_size=dict_size)

        # Compress with py7zr
        return compress_7z_py(
            path_in=path_in,
            path_out=path_out)

    # Error occurred, None returned
    return logger.exception(f'Unable to compress file: {path_in.name}')


def compress_7z_all(
    path_in: Path,
    path_out: Path = None,
    word_size: WordSize = WordSize.WS16,
    dict_size: DictionarySize = DictionarySize.DS1536
) -> None:
    """Compress every file inside `path_in` directory as 7z archives, then output
    those archives in the `path_out`.

    Args:
        path_in: Directory containing files to compress.
        path_out: Directory to place the archives. Use a subdirectory 'compressed' if not provided.
        word_size: Word size value to use for the compression.
        dict_size: Dictionary size value to use for the compression.
    """
    # Use "compressed" subdirectory if not provided, ensure output directory exists
    path_out = path_out or Path(path_in, '.compressed')
    path_out.mkdir(mode=777, parents=True, exist_ok=True)

    # Get a list of all .psd files in the directory
    files = [
        Path(path_in, n) for n in os.listdir(path_in)
        if Path(path_in, n).is_file()]

    # Compress each file
    with tqdm(total=len(files), desc="Compressing files", unit="file") as pbar:
        for f in files:
            p = (path_out / f.name).with_suffix('.7z')
            pbar.set_description(f.name)
            compress_7z(
                path_in=f,
                path_out=p,
                word_size=word_size,
                dict_size=dict_size)
            pbar.update()


"""
* Decompression Utils
"""


def unpack_zip(path: Path) -> None:
    """Unpack target 'zip' archive.

    Args:
        path: Path to the archive.

    Raises:
        FileNotFoundError: If archive couldn't be located.
    """
    if not path.is_file():
        raise FileNotFoundError(f'Archive not found: {str(path)}')
    with zipfile.ZipFile(path) as z:
        z.extractall(path=path.parent)


def unpack_gz(path: Path) -> None:
    """Unpack target 'gz' archive.

    Args:
        path: Path to the archive.

    Raises:
        FileNotFoundError: If archive couldn't be located.
    """
    if not path.is_file():
        raise FileNotFoundError(f'Archive not found: {str(path)}')
    output = path.parent / path.name[:-3]
    with gzip.open(path) as z:
        with open(output, 'wb') as f:
            shutil.copyfileobj(z, f)


def unpack_xz(path: Path) -> None:
    """Unpack target 'xz' archive.

    Args:
        path: Path to the archive.

    Raises:
        FileNotFoundError: If archive couldn't be located.
    """
    if not path.is_file():
        raise FileNotFoundError(f'Archive not found: {str(path)}')
    output = path.parent / path.name[:-3]
    with lzma.open(path) as z:
        with open(output, 'wb') as f:
            shutil.copyfileobj(z, f)


def unpack_bz2(path: Path) -> None:
    """Unpack target 'bz2' archive.

    Args:
        path: Path to the archive.

    Raises:
        FileNotFoundError: If archive couldn't be located.
    """
    if not path.is_file():
        raise FileNotFoundError(f'Archive not found: {str(path)}')
    output = path.parent / path.name[:-3]
    with bz2.open(path) as z:
        with open(output, 'wb') as f:
            shutil.copyfileobj(z, f)


def unpack_7z(path: Path) -> None:
    """Unpack target '7z' archive.

    Args:
        path: Path to the archive.

    Raises:
        FileNotFoundError: If archive couldn't be located.
    """
    if not path.is_file():
        raise FileNotFoundError(f'Archive not found: {str(path)}')
    with py7zr.SevenZipFile(path, 'r') as z:
        z.extractall(path=path.parent)


def unpack_tar(path: Path, mode: str = 'gz'):
    """Unpack target 'tar' archive of a given type.

    Args:
        path: Path to the archive.
        mode: Mode to use when unpacking, i.e. type of archive it is (gz, xz, etc). Defaults to `gz`.

    Raises:
        FileNotFoundError: If archive couldn't be located.
    """
    if not path.is_file():
        raise FileNotFoundError(f'Archive not found: {str(path)}')
    with tarfile.open(path, f'r:{mode}') as z:
        z.extractall(path=path.parent)


def unpack_tar_gz(path: Path) -> None:
    """Shorthand function for `unpack_tar` targeting a `tar.gz` archive."""
    unpack_tar(path, mode='gz')


def unpack_tar_xz(path: Path) -> None:
    """Shorthand function for `unpack_tar` targeting a `tar.xz` archive."""
    unpack_tar(path, mode='xz')


def unpack_tar_bz2(path: Path) -> None:
    """Shorthand function for `unpack_tar` targeting a `tar.bz2` archive."""
    unpack_tar(path, mode='bz2')


def unpack_archive(path: Path, remove: bool = True, thread_lock: Optional[Lock] = None) -> None:
    """Unpack an archive using the correct methodology based on its extension.

    Args:
        path: Path to the archive.
        remove: Whether to remove the archive after unpacking.
        thread_lock: Optional Lock object used to prevent concurrent unpacking, will use
            default Lock object if not provided.

    Raises:
        FileNotFoundError: If archive couldn't be located.
    """
    action_map: dict[str, Callable] = {
        ArchType.Zip: unpack_zip,
        ArchType.GZip: unpack_gz,
        ArchType.XZip: unpack_xz,
        ArchType.TarGZip: unpack_tar_gz,
        ArchType.TarXZip: unpack_tar_xz,
        ArchType.SevenZip: unpack_7z,
        ArchType.TarSevenZip: unpack_7z
    }
    if path.suffix not in action_map:
        return
    action = action_map[path.suffix]
    if thread_lock is None:
        thread_lock = ARCHIVE_LOCK
    with thread_lock:
        _ = action(path)
    if remove:
        os.remove(path)
    gc.collect()



"""
* Test Utilities
* Generalized utilities for performing tests.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
from omnitils.test.bench import *



"""
* Google Drive Utilities
* Generalized utilities for accessing Google Drive files or Google Drive API endpoints.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
from contextlib import suppress
from dataclasses import dataclass
import os
from pathlib import Path
import re
import shutil
from typing import Optional, Callable, TypedDict, NotRequired, Union

# Third Party Imports
from loguru import logger
import requests
import yarl
from requests import Session, Response

# Local Imports
from omnitils.fetch import (
    request_header_default,
    get_new_session,
    chunk_size_default,
    download_file_from_response)
from omnitils.files import mkdir_full_perms, get_temporary_file, dump_data_file
from omnitils.strings import decode_url

"""
* Enums
"""


@dataclass
class GoogleReg:
    """Defined Google URL regex patterns."""
    URL: re.Pattern = re.compile(r'"downloadUrl":"([^"]+)')
    FORM: re.Pattern = re.compile(r'id="download-form" action="(.+?)"')
    EXPORT: re.Pattern = re.compile(r'href="(/uc\?export=download[^"]+)')
    ERROR: re.Pattern = re.compile(r'<p class="uc-error-subcaption">(.*)</p>')


"""
* Types
"""


class GoogleDriveMetadata(TypedDict):
    """Relevant metadata for a file hosted on Google Drive."""
    description: NotRequired[str]
    name: str
    size: int


"""
* Google Drive Utils
"""


def gdrive_get_confirmation_url(contents: str) -> yarl.URL:
    """Get the correct URL for downloading a Google Drive file from a potential
        confirmation page contents.

    Args:
        contents: Google Drive page contents.

    Returns:
        URL object targeting the hosted file.
    """
    for line in contents.splitlines():
        if m := GoogleReg.EXPORT.search(line):
            # Google Docs URL
            return decode_url(f'https://docs.google.com{m.groups()[0]}')
        if m := GoogleReg.FORM.search(line):
            # Download URL from Form
            return decode_url(m.groups()[0])
        if m := GoogleReg.URL.search(line):
            # Download URL from JSON
            return decode_url(m.groups()[0])
        if m := GoogleReg.ERROR.search(line):
            # Error Returned
            raise OSError(m.groups()[0])
    raise OSError(
        "Google Drive file has been made private or has reached its daily request limit.")


def gdrive_process_url(
    url: Union[str, yarl.URL],
    sess: Optional[Session] = None,
    headers: Optional[dict] = None,
    path_cookies: Optional[Path] = None
) -> Optional[tuple[Session, Response]]:
    """Tests a Gdrive file URL to ensure it is the absolute download URL. If it isn't,
        attempt to redirect to the absolute URL based on Google Drive confirmation. Return a valid session and
        response object if successful, otherwise None.

    Args:
        url: Url to Google Drive asset.
        sess: Session object to use for requests, create one if not provided.
        headers: Headers to pass when creating a new Session object, if provided.
        path_cookies: Path to cookies file to load saved cookies from, will skip cookies if not provided.

    Returns:
        A tuple containing the Session object and Response from correct URL, if successful. Returns None if
            a working URL couldn't be established or Google Drive denies access to the file.
    """
    # Ensure session object, then initialize request
    if not sess:
        sess = get_new_session(
            headers=headers, stream=True, path_cookies=path_cookies)
    res = sess.get(url)

    # Update cookies
    if path_cookies:
        gdrive_update_cookies(sess, path_cookies)

    # Is this the right file?
    if "Content-Disposition" in res.headers:
        return sess, res

    # Try again with updated URL from confirmation
    try:
        return gdrive_process_url(
            url=gdrive_get_confirmation_url(res.text),
            sess=sess,
            headers=headers,
            path_cookies=path_cookies)
    except Exception as e:
        logger.error(e), res.close(), sess.close()
        return logger.error(f'Google Drive denied access to the file!')


"""
* Manage Gdrive Sessions
"""


def gdrive_update_cookies(sess: Session, path_cookies: Path) -> None:
    """Update cookies file using Gdrive session.

    Args:
        sess: Session object to pull cookies from.
        path_cookies: Path to local cookies file.
    """
    dump_data_file(
        obj=[
            (k, v) for k, v in sess.cookies.items()
            if not k.startswith("download_warning_")
        ],
        path=path_cookies)


"""
* Fetching Metadata
"""


def gdrive_get_metadata(
    file_id: str,
    api_key: str,
    header: Optional[dict] = None
) -> Optional[GoogleDriveMetadata]:
    """Get the metadata of a given template file.

    Args:
        file_id: ID of the Google Drive file.
        api_key: Google Drive API key.
        header: Optional header to pass with request.

    Returns:
        Metadata of the Google Drive file.
    """
    if header is None:
        header = request_header_default.copy()
    with suppress(Exception):
        with requests.get(
            f"https://www.googleapis.com/drive/v3/files/{file_id}",
            headers=header,
            params={
                'alt': 'json',
                'fields': 'description,name,size',
                'key': api_key}
        ) as req:
            if not req.status_code == 200:
                return
            result = req.json()
            if 'name' in result and 'size' in result:
                return result

    # Request was unsuccessful
    return


"""
* Downloading Files
"""


def gdrive_download_file(
    url: Union[yarl.URL, str],
    path: Path,
    callback: Optional[Callable] = None,
    headers: Optional[dict] = None,
    path_cookies: Optional[Path] = None,
    allow_resume: bool = True,
    chunk_size: int = chunk_size_default
) -> Optional[Path]:
    """Download a file from Google Drive using its file ID.

    Note:
        Ideal file URL: https://drive.google.com/uc?id={file_id}

    Args:
        url: Google Drive file ID.
        path: Path to save downloaded file.
        callback: Function to call on each chunk downloaded.
        headers: Headers object to pass with request, uses default if not provided.
        path_cookies: Path to cookies file, don't use cookies if not provided.
        allow_resume: Whether to allow resuming a previous download.
        chunk_size: Chunk size in bytes when downloading the file with `iter_content`. Defaults to 8MB.

    Returns:
        Path to the downloaded file if successful, otherwise None.
    """
    # Ensure path and load a temporary file
    mkdir_full_perms(path.parent)
    file = get_temporary_file(
        path=path, ext='.drive', allow_existing=allow_resume)
    size = file.stat().st_size

    # Add range header if file is partially downloaded
    headers = headers or request_header_default.copy()
    if size > 0:
        headers["Range"] = f"bytes={str(size)}-"

    # Attempt to create a session and request from URL
    check = gdrive_process_url(
        url=url, headers=headers, path_cookies=path_cookies)
    if not check:
        return logger.error(
            f'Google Drive download failed!\n'
            f'{path.name} | {url}')
    sess, res = check

    # Attempt to download the file
    try:
        result = download_file_from_response(
            response=res,
            path=file,
            callback=callback,
            chunk_size=chunk_size)
    except Exception as e:
        # Exception occurred
        logger.error(e)
        result = None
    if result is None:
        # Download failed
        res.close(), sess.close()
        return logger.error(
            f'Google Drive download failed!\n'
            f'{path.name} | {url}')

    # Rename temporary file
    if not os.path.samefile(file, path):
        shutil.move(file, path)

    # Close session and return
    res.close(), sess.close()
    return path



"""
* Image Utilities
* Various utilities for working with images.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
from omnitils.img._core import *



"""
* Logging Utilities
* Generalized utilities for loggers and logging output.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
import copy
from typing import Any, Callable, Optional
import sys

# Third Party Imports
from loguru import logger as loguru_logger
from loguru._logger import Logger

# Reset the loguru logger
loguru_logger.remove()
logger = copy.deepcopy(loguru_logger)

"""
* Default Logging Definitions
"""

# Pre-defined component
COMPONENT_TIME = "{time:YY}.{time:MM}.{time:DD} {time:HH}:{time:mm}:{time:ss}"
COMPONENT_LEVEL = "<lvl>{level: <8}</lvl>"
COMPONENT_LOCATION = "<pos>[{location}]</pos>"
COMPONENT_MESSAGE = "<msg>{message}</msg>"
COMPONENT_EXCEPTION = "<lr>{exception}</lr>"

# Common tags
TAGS_DEFAULT = ('<w>', '</w>')

# Pre-defined colors
LEVEL_COLORS = {
    'DEBUG': dict(
        lvl=['lm'],
        msg=['m'],
        pos=['lc']),
    'INFO': dict(
        lvl=['le'],
        msg=['e'],
        pos=['lc']),
    'SUCCESS': dict(
        lvl=['lg'],
        msg=['g'],
        pos=['lc']),
    'WARNING': dict(
        lvl=['b', 'ly'],
        msg=['y'],
        pos=['ly']),
    'ERROR': dict(
        lvl=['lr', 'b'],
        msg=['r'],
        pos=['lr']),
    'CRITICAL': dict(
        lvl=['lw', 'bg 130,0,0', 'b'],
        msg=['lr', 'b'],
        pos=['lr'])
}

"""
* Logging Handlers
"""


def colorize_log_format(log_fmt: str, log_level: str) -> str:
    """An internal utility function called by a handler to colorize custom tags in a log format string.

    Args:
        log_fmt: Log format string.
        log_level: Level of current log to be formatted.

    Returns:
        str: Colorized log format string.
    """

    # Get log level colors
    log_colors = LEVEL_COLORS.get(log_level) or LEVEL_COLORS['DEBUG']

    # Replace each tag
    for name, tag_group in log_colors.items():
        left, right = '', ''
        for tag in tag_group:

            # Open and close defined
            if isinstance(tag, (tuple, list)):
                if len(tag) > 1:
                    tag_open, tag_close = tag
                    left = f'{left}<{tag_open}>'
                    right = f'</{tag_close}>{right}'
                    continue
                elif len(tag) == 1:
                    tag = tag[0]

            # One definition
            left += f'<{tag}>'
            right = f'</{tag}>{right}'

        # Replace each opening and closing tag
        log_fmt = log_fmt.replace(f'<{name}>', left).replace(f'</{name}>', right)
    return log_fmt


def formatting_handler(record: dict[str, Any]) -> str:
    """An internal function used to return a granular format to the Loguru logger object.

    Args:
        record: Loguru logger record.

    Returns:
        str: Format to be used for logger output.
    """

    # Establish base values
    _level = record['level'].name
    _extra = record.get('extra', {})
    _msg = record.get('message', '')
    _exception = record.get('exception')
    _fmt = ''

    # Check if exception exists and should be shown
    _show_time = bool(_extra.get('show_time', True))
    _show_level = bool(_extra.get('show_level', True))
    _show_location = bool(_extra.get('show_location', bool(_level in ['WARNING', 'ERROR', 'CRITICAL'])))
    _show_message = bool(_extra.get('show_message', True))
    _show_exception = bool(
        _extra.get('show_exception', True) and
        hasattr(_exception, 'traceback') and
        _exception.traceback is not None)

    # Check for alternate separator
    _sep = _extra.get('separator', ' <b>|</b> ')

    # Check for a continuation line
    terminator = '\n'
    if not _show_exception:
        if bool(_extra.get('await_more', False)):
            terminator = ''
        if bool(_extra.get('add_more', False)):
            return colorize_log_format(
                COMPONENT_MESSAGE.replace('{message}', _msg),
                _level
            ) + terminator

    def _add_component(_log_fmt, _component, _separator: Optional[str] = _sep) -> str:
        """Adds a component to a provided log format, with a separator."""
        if _log_fmt == '':
            return _component
        return _log_fmt + _separator + _component

    def _wrap_main_tags(_log_fmt: str):
        """Wraps the log format in the main surrounding tags."""
        _L, _R = TAGS_DEFAULT
        if _main_tags := _extra.get('main_tags'):
            if isinstance(_main_tags, tuple) and len(_main_tags) == 2:
                _L, _R = _main_tags
        return _L + _log_fmt + _R + terminator

    # Add time
    if _show_time:
        _fmt = _add_component(_fmt, COMPONENT_TIME)

    # Add level
    if _show_level:
        _fmt = _add_component(_fmt, COMPONENT_LEVEL)

    # Add location
    if _show_location:
        _this_func = '' if '<' in (record.get('function') or '') else '.{function}'
        _this_location = '{module}' + _this_func + '.{line}'
        _component = COMPONENT_LOCATION.replace(
            '{location}', _this_location)
        _fmt = _add_component(_fmt, _component)

    # Add message
    if _show_message:
        _fmt = _add_component(_fmt, COMPONENT_MESSAGE.replace('{message}', _msg))

    if _show_exception:
        _fmt = _add_component(_fmt, COMPONENT_EXCEPTION, '\n')

    # Inject color tags, wrap, terminate, and return
    return _wrap_main_tags(colorize_log_format(_fmt, _level))


# Pre-defined handlers
HANDLER_DEFAULT = dict(
    sink=sys.stderr,
    format=formatting_handler,
    backtrace=True,
    diagnose=True,
    level='DEBUG'
)


"""
* Configure Logger
"""


def reconfigure_logger(
    obj: Logger = logger,
    handlers: Optional[list[dict[str, Any]]] = None,
    **kwargs
) -> Logger:
    """Returns a configured loguru logger object."""
    if not handlers:
        handlers = [HANDLER_DEFAULT]
    obj.configure(handlers=handlers, **kwargs)
    return obj


def get_logger(handlers: Optional[list[dict[str, Any]]] = None, **kwargs) -> Logger:
    """Return a unique loguru logger object with optional provided handlers.

    Args:
        handlers: A list of handler dict definitions, otherwise will use the default handler.

    Returns:
        A unique loguru logger object.
    """
    _logger: Logger = copy.deepcopy(loguru_logger)
    _handlers: list[dict[str, Any]] = []

    # Handlers provided
    if handlers is not None:
        for n in handlers:
            _base = HANDLER_DEFAULT.copy()
            _base.update(n)
            _handlers.append(_base)
        return reconfigure_logger(_logger, _handlers, **kwargs)
    return reconfigure_logger(_logger, **kwargs)


"""
* Context Managers
"""


class TemporaryLogger:
    def __init__(
        self,
        handlers: Optional[list[dict[str, Any]]] = None
    ):
        self.logger: Logger = copy.deepcopy(loguru_logger)
        self._handlers: list[dict[str, Any]] = []
        if handlers is not None:
            for n in handlers:
                _base = HANDLER_DEFAULT.copy()
                _base.update(n)
                self._handlers.append(_base)
        else:
            self._handlers.append(HANDLER_DEFAULT.copy())

    def __enter__(self) -> Logger:

        # Add handlers
        self.logger = reconfigure_logger(self.logger, handlers=self._handlers)
        return self.logger

    def __exit__(self, exc_type, exc_value, traceback):

        # Remove the new sink
        del self.logger


class LogResults:
    """Context manager to log the success, failure, and/or exceptions within the context being executed. A success
    message (if provided) will be logged if the context completes without raising an exception, otherwise an error
    message (if provided) will be logged, and/or the exception itself if requested.

    Args:
        on_failure: Message to log if context encounters an exception, if provided.
        on_success: Message to log if context executes successfully, if provided.
        reraise: Whether to reraise an encountered exception after it is caught and logged.
        log_trace: Whether to log the traceback of any exception that occurs.
        log_obj: Logger object to use, defaults to the main omnitils logger.
    """

    def __init__(
        self,
        on_failure: Optional[str] = None,
        on_success: Optional[str] = None,
        reraise: bool = False,
        log_trace: bool = True,
        log_obj: Logger = logger
    ):
        self._on_failure = on_failure
        self._on_success = on_success
        self._reraise = reraise
        self._log_trace = log_trace
        self._logger = log_obj

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):

        # Log error message and/or traceback
        if exc_type:
            if self._log_trace:
                if self._on_failure is None:
                    self._on_failure = 'The following exception occurred:'
                self._logger.exception(self._on_failure)
            elif self._on_failure:
                self._logger.error(self._on_failure, show_location=False)
            # Re-raise the exception
            return False if self._reraise else True

        # Context executed without errors
        if self._on_success:
            self._logger.success(self._on_success)


"""
* Decorators
"""


def log_test_result(
    on_failure: Optional[str] = None,
    on_success: Optional[str] = None,
    on_failure_return: Optional[Any] = None,
    reraise: bool = False,
    log_trace: bool = True,
    log_obj: Logger = logger
) -> Callable:
    """Catch and log any errors, otherwise log a success.

    Args:
        on_failure: Message to be logged if decorated function encounters an exception.
        on_success: Message to be logged if decorated function executes successfully.
        on_failure_return: Return value if exception occurs, defaults to None.
        reraise: Whether to reraise an encountered exception after it is caught and logged.
        log_trace: Whether to log the traceback of any exception that occurs.
        log_obj: Logger object to use, defaults to the main omnitils logger. Must be a loguru logger object.

    Returns:
        Wrapped function.
    """
    def decorator(func: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            try:
                with LogResults(
                    on_success=on_success,
                    on_failure=on_failure,
                    reraise=True,
                    log_trace=log_trace,
                    log_obj=log_obj
                ):
                    return func(*args, **kwargs)
            except Exception as e:
                if reraise:
                    raise e
                return on_failure_return
        return wrapper
    return decorator


# Configure main logger object
reconfigure_logger()



"""
* Data File Utilities
* Generalized utilities for working with data files, i.e. JSON, YAML, etc.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
import json
from contextlib import suppress
from pathlib import Path
from typing import Optional, TypedDict, Callable, Union
from threading import Lock

# Third Party Imports
from yaml import (
    load as yaml_load,
    dump as yaml_dump,
    Loader as yamlLoader,
    Dumper as yamlDumper)
from tomlkit import dump as toml_dump, load as toml_load

"""
* Types
"""


class DataFileType (TypedDict):
    """Data file type (json, toml, yaml, etc)."""
    load: Callable
    dump: Callable
    load_kw: dict[str, Union[Callable, bool, str]]
    dump_kw: dict[str, Union[Callable, bool, str]]


"""Data File: TOML (.toml) data type."""
DataFileTOML = DataFileType(
    load=toml_load,
    dump=toml_dump,
    load_kw={},
    dump_kw={'sort_keys': True})

"""Data File: YAML (.yaml) data type."""
DataFileYAML = DataFileType(
    load=yaml_load,
    dump=yaml_dump,
    load_kw={
        'Loader': yamlLoader},
    dump_kw={
        'allow_unicode': True,
        'Dumper': yamlDumper,
        'sort_keys': True,
        'indent': 2,
    })

"""Data File: JSON (.json) data type."""
DataFileJSON = DataFileType(
    load=json.load,
    dump=json.dump,
    load_kw={},
    dump_kw={
        'sort_keys': True,
        'indent': 2,
        'ensure_ascii': False
    })


"""
* Constants
"""

# File util locking mechanism
util_file_lock = Lock()

# Data types alias map
data_types: dict[str, DataFileType] = {
    '.toml': DataFileTOML,
    '.yaml': DataFileYAML,
    '.yml': DataFileYAML,
    '.json': DataFileJSON,
}
supported_data_types = tuple(data_types.keys())

"""
* Funcs
"""


def validate_data_type(path: Path) -> None:
    """Checks if a data file matches a supported data file type.

    Args:
        path: Path to the data file.

    Raises:
        ValueError: If data file type not supported.
    """
    # Check if data file is a supported data type
    if path.suffix.lower() not in supported_data_types:
        raise ValueError("Data file provided does not match a supported data file type.\n"
                         f"Types supported: {', '.join(supported_data_types)}\n"
                         f"Type received: {path.suffix}")


def validate_data_file(path: Path) -> None:
    """Checks if a data file exists and is a valid data file type. Raises an exception if validation fails.

    Args:
        path: Path to the data file.

    Raises:
        FileNotFoundError: If data file does not exist.
        ValueError: If data file type not supported.
    """
    # Check if file exists
    if not path.is_file():
        raise FileNotFoundError(f"Data file does not exist:\n{str(path)}")
    validate_data_type(path)


def load_data_file(
    path: Path,
    config: Optional[dict] = None
) -> Union[list, dict, tuple, set]:
    """Load data  object from a data file.

    Args:
        path: Path to the data file to be loaded.
        config: Dict data to modify DataFileType configuration for this data load procedure.

    Returns:
        Data object such as dict, list, tuple, set, etc.

    Raises:
        FileNotFoundError: If data file does not exist.
        ValueError: If data file type not supported.
        OSError: If loading data file fails.
    """
    # Check if data file is valid
    validate_data_file(path)

    # Pull the parser and insert user config into kwargs
    parser: DataFileType = data_types.get(path.suffix.lower(), {}).copy()
    if config:
        parser['load_kw'].update(config)

    # Attempt to load data
    with util_file_lock, suppress(Exception), open(path, 'r', encoding='utf-8') as f:
        data = parser['load'](f, **parser['load_kw']) or {}
        return data
    raise OSError(f"Unable to load data from data file:\n{str(path)}")


def dump_data_file(
    obj: Union[list, dict, tuple, set],
    path: Path,
    config: Optional[dict] = None
) -> None:
    """Dump data object to a data file.

    Args:
        obj: Iterable or dict object to save to data file.
        path: Path to the data file to be dumped.
        config: Dict data to modify DataFileType configuration for this data dump procedure.

    Raises:
        FileNotFoundError: If data file does not exist.
        ValueError: If data file type not supported.
        OSError: If dumping to data file fails.
    """
    # Check if data file is valid
    validate_data_type(path)

    # Pull the parser and insert user config into kwargs
    parser: DataFileType = data_types.get(path.suffix.lower(), {}).copy()
    if config:
        parser['dump_kw'].update(config)

    # Attempt to dump data
    with suppress(Exception), util_file_lock, open(path, 'w', encoding='utf-8') as f:
        parser['dump'](obj, f, **parser['dump_kw'])
        return
    raise OSError(f"Unable to dump data to data file:\n{str(path)}")


"""
* Project Files
"""


def get_project_version(path: Path) -> str:
    """Returns the version string stored in the root project file.

    Args:
        path: Path to the root project file.

    Returns:
        Current version string.

    Raises:
        FileNotFoundError: If project file does not exist.
        ValueError: If project file type not supported.
        OSError: If project file fails to load.
    """
    project = load_data_file(path)
    return project.get('tool', {}).get('poetry', {}).get('version', '1.0.0')



"""
* CLI Commands: Testing
* Internal testing utility commands.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
import os
import shutil
from pathlib import Path

# Third Party Imports
import click
from loguru import logger

# Local Imports
from omnitils.api.github import (
    gh_download_repository,
    gh_download_directory_files)

# Test resources
temp_dir = Path(os.getcwd(), 'temp')

"""
* Test Group: Github
"""


@click.command()
def test_gh_download_repository():
    """Tests the use of `omnitils.fetch.gh_download_repository`."""

    # Setup test
    # Todo: Create context handler for creating and disposing of temporary directory
    if not temp_dir.is_dir():
        os.mkdir(temp_dir)

    # Perform test
    extracted = gh_download_repository(
        user='Investigamer',
        repo='omnitils',
        path=temp_dir)
    check_file = extracted / 'poetry.lock'
    try:
        # Check repo directory
        assert extracted.is_dir()
        try:
            # Check test file
            assert check_file.is_file()
            logger.success('Test passed!')
        except AssertionError:
            logger.error('Test file missing from downloaded repo!')
        shutil.rmtree(extracted)
    except AssertionError:
        logger.error('Repo directory not downloaded!')

    # Test completed
    shutil.rmtree(temp_dir)


@click.command()
def test_gh_download_directory_files():
    """Tests the user of `omnitils.fetch.gh_download_directory_files`."""

    # Setup test
    if not temp_dir.is_dir():
        os.mkdir(temp_dir)

    # Perform test
    files_downloaded = gh_download_directory_files(
        user='Investigamer',
        repo='omnitils',
        repo_dir='omnitils',
        path=temp_dir)
    try:
        # Check repo directory
        assert len(files_downloaded) > 0
        [os.remove(n) for n in files_downloaded]
        logger.success('Test passed!')
    except AssertionError:
        return logger.error('No files downloaded from repo!')

    # Test completed
    shutil.rmtree(temp_dir)


@click.group(
    commands={
        'files': test_gh_download_directory_files,
        'repo': test_gh_download_repository
    }
)
def github_group():
    """Command group for testing the `fetch.github` module."""
    pass


"""
* Main Testing Group
"""


@click.group(commands={'fetch-gh': github_group})
def testing_group():
    """Command group for performing tests."""
    pass



"""
* Download Utilities
* Generalized utilities for downloading files.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
import os
from contextlib import suppress
from typing import Union, Callable, Optional

# Third Party Imports
import requests
from requests import Response
from yarl import URL

# Local Imports
from omnitils.fetch._core import request_header_default, chunk_size_default


"""
* Working With Headers
"""


def check_resume_file(path: Union[str, os.PathLike], headers: dict[str, str]) -> bool:
    """Checks if a file download can be resumed by looking at the file and headers.

    Args:
        path: Path to the file being downloaded.
        headers: Headers being passed in the download request.

    Returns:
        True if resuming, otherwise False.
    """
    # Check for a range header with bytes spec
    if headers.get('Range') is not None and os.path.exists(path):
        if 'bytes=' in headers['Range']:
            try:
                # Check that range and size is equivalent
                status = headers['Range'][6:].split('-')[0]
                if int(status) == os.path.getsize(path):
                    return True
            except (IndexError, KeyError, TypeError):
                return False
    return False


def estimate_content_length(response: Response, default: int = 0) -> int:
    """Attempt to get the value of the Content-Length header from a Response. Return default value if not found.

    Args:
        response: Response object.
        default: Default value to return if no Content-Length header can't be parsed.

    Returns:
        Content-Length value from Response headers.
    """
    total = response.headers.get('Content-Length', default)
    with suppress(Exception):
        if isinstance(total, str):
            total = int(''.join(n for n in total if n.isdigit()) or default)
        if isinstance(total, int):
            return total
    return default


"""
* Download Requests
"""


def download_file(
    url: Union[str, URL],
    path: Union[str, os.PathLike],
    header: Optional[dict] = None,
    callback: Optional[Callable] = None,
    chunk_size: int = chunk_size_default,
) -> Union[str, os.PathLike]:
    """Download a file in chunks from url and save to path, executing a callback
        after each chunk if provided.

    Args:
        url: URL where the file is hosted.
        path: Path to save the file to.
        header: Header object to pass with request, uses default if not provided.
        callback: Callback to execute after each chunk is written. Passes
            number of bytes written (int) and number of bytes total (int).
        chunk_size: Chunk size in bytes to download while streaming the file.

    Returns:
        Path to the saved file, if successful.

    Raises:
        RequestException: If request is unsuccessful.
        FileExistsError: If file already exists and cannot be overwritten.
    """
    has_callback = callback is not None
    header = header or request_header_default.copy()
    write_mode = 'ab' if check_resume_file(path, header) else 'wb'
    with requests.get(url, headers=header, stream=True) as r:
        r.raise_for_status()

        # Get file size total
        total = estimate_content_length(r) if has_callback else 0

        # Write the file in chunks
        with open(path, write_mode) as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                if not chunk:
                    raise OSError('Bad chunk detected, likely a truncated stream!')
                f.write(chunk)

                # Execute callback if provided
                if has_callback:
                    current = f.tell()
                    callback(current, max(current, total))

    # Return path
    return path


def download_file_from_response(
    response: Response,
    path: Union[str, os.PathLike],
    callback: Optional[Callable[[int, int], None]] = None,
    chunk_size: int = chunk_size_default,
) -> Union[str, os.PathLike]:
    """Download a file in chunks using a given Request and optional Session, executing a callback
        after each chunk if provided.

    Args:
        response: Response stream to download file from.
        path: Path to save the file to.
        callback: Optional callback to execute after each chunk is written. Passes
            number of bytes written (int) and number of bytes total (int).
        chunk_size: Chunk size in bytes to download over each `iter_content` iteration, defaults to 8MB.

    Returns:
        Path to the saved file, if successful.

    Raises:
        RequestException: If request is unsuccessful.
        FileExistsError: If file already exists and cannot be overwritten.
    """
    response.raise_for_status()
    has_callback = callback is not None

    # Get file size total
    total = estimate_content_length(response) if has_callback else 0
    write_mode = 'ab' if check_resume_file(path, response.headers) else 'wb'

    # Write the file in chunks
    with open(path, write_mode) as f:
        for chunk in response.iter_content(chunk_size=chunk_size):
            # Check for bad chunks
            if not chunk:
                raise OSError('Bad chunk detected, likely a truncated stream!')
            f.write(chunk)

            # Execute callback if provided
            if has_callback:
                current = f.tell()
                callback(current, max(current, total))

    # Return path
    return path



"""
* Metaclass Utilities
* General utility meta classes.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""

"""
* Meta-class Utils
"""


class Singleton(type):
    """Maintains a single instance of any child class to return for any subsequent calls."""
    _instances: dict = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



"""
* Benchmarking Utilities
* Generalized utilities for performing benchmarking tests.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
import functools
# Standard Library Imports
from time import perf_counter
from typing import Optional, Callable, Any

# Third Party
from omnitils.logs import logger as _logger

# Local Imports
from omnitils.schema import Schema

"""
* Types
"""


class BenchmarkResult(Schema):
    """Represents a benchmarking result."""
    value: Any
    average: float
    times: list = []
    name: str


"""
* Util Decorators
"""


class time_function:
    """Print the execution time in seconds of any decorated function.

    Args:
        msg_or_func:
            * If the decorator is not called, this serves as the decorated function which is cached for
                use in the wrapper.
            * If the decorator is called, you may provide a string used to format the logger output which
                that prints the execution time. This string supports optional format variables {f} and {t}.
                "f" will be formatted with the function name, "t" will be formatted with the execution time.
            * Example: "Function `{f}` completed in {t:.2f} seconds."
        logger: Logging object to use.
    """

    def __init__(self, msg_or_func: Optional[str | Callable] = None, logger: Any = None):
        self._func = None
        self._msg = 'Function `{f}` completed in {t:.4f} seconds.'
        self._logger = logger or _logger

        # Use argument as function or log format
        if callable(msg_or_func):
            self._func = msg_or_func
        elif isinstance(msg_or_func, str):
            self._msg = msg_or_func

    def __call__(
        self,
        *args,
        **kw
    ) -> Any:
        """Print the execution time in seconds of any decorated function.

        Returns:
            Wrapped function.
        """
        is_wrapped: bool = bool(self._func is None)
        _func: Callable = [*args].pop() if is_wrapped else self._func
        _func_name: str = _func.__name__

        @functools.wraps(_func)
        def wrapper(*_args, **_kwargs):
            """Wrapped function."""
            s = perf_counter()
            result = _func(*_args, **_kwargs)
            self._logger.info(
                self._msg.format(
                    f=_func_name,
                    t=(perf_counter() - s)))
            return result

        # Return result or wrapped function
        return wrapper if is_wrapped else wrapper(*args, **kw)


"""
* Util Funcs
"""


def benchmark_funcs(
    funcs: list[tuple[Callable, list[Any]]],
    iterations: int = 1000,
    reset_func: Optional[Callable] = None,
    logger: Any = None
) -> None:
    """Test the execution time of a new function against an older function.

    Args:
        funcs: List of tuples containing a func to test and args to pass to it.
        iterations: Number of calls to each function to perform, must be higher than 1.
        reset_func: Optional function to call to reset app state between actions.
        logger: Logging object to use.
    """
    logger = logger or _logger

    # Skip if no funcs provided
    if not funcs:
        return logger.critical('No functions provided for benchmarking.')
    if iterations < 1:
        return logger.critical('Iterations must be 1 or higher.')

    # Test configuration
    results: list[BenchmarkResult] = []
    best: float = 0

    # Test each function
    for func, args in funcs:
        times, value = [], None

        # Track the execution time across iterations
        for i in range(iterations):
            s = perf_counter()
            try:
                value = func(*args)
            except Exception as e:
                logger.error(f'Encountered an error in function "{func.__name__}"!')
                return logger.exception(e)
            times.append(perf_counter()-s)
            if reset_func:
                reset_func()

        # Append result and value
        results.append(
            BenchmarkResult(
                value=value,
                average=sum(times)/len(times),
                times=times,
                name=func.__name__))

    # Report results
    for i, r in enumerate(sorted(results, key=lambda item: item.average)):

        # Best result
        if i == 0:
            best = r.average
            logger.success(f'{r.name}: {r.average}')
            continue

        # Slower results
        pct_slower = round(((r.average - best) / ((r.average + best) / 2)) * 100, 2)
        logger.warning(f"{r.name}: {r.average} ({pct_slower}% Slower)")

    # Check if all values match
    if not results:
        return logger.warning('No results were found to compare.')
    if all(n.value == results[0].value for n in results):
        return logger.success('Values are identical!')

    # Values don't match
    logger.error("Values don't appear to be identical! See values below.")
    [logger.info(n.value) for n in results]
    return



"""
* Folder Utilities
* Generalized utilities for working with folders/directories.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
import os
import random
import shutil
from pathlib import Path
from typing import Iterator, Optional

"""
* Creating Directories
"""


def mkdir_full_perms(path: Path) -> Path:
    """Uses a given path to create a directory if it doesn't exist, with full permissions.

    Args:
        path: Path to the directory.

    Returns:
        The director provided.
    """
    if path.is_dir():
        return path
    os.umask(0)
    path.mkdir(parents=True, exist_ok=True)
    return path


"""
* Directory Information
"""


def get_subdirs(path: Path) -> Iterator[Path]:
    """Yields each subdirectory of a given folder.

    Args:
        path: Path to the folder to iterate over.

    Yields:
        A subdirectory of the given folder.
    """
    for dir_path, dir_names, filenames in os.walk(path):
        for dirname in dir_names:
            yield Path(dir_path) / dirname


def is_dir_empty(path: Path) -> bool:
    """Checks if the given directory is empty.

    Args:
        path: Path to the directory.

    Returns:
        True if the directory is empty, otherwise False.

    Raises:
        FileNotFoundError: If the given directory doesn't exist.
    """
    if not path.is_dir():
        raise FileNotFoundError('Directory does not exist: {}'.format(path))
    return not os.listdir(path)


"""
* Moving Directories and Contents
"""


def recur_move_dir(source: Path, target: Path, overwrite: bool = True) -> None:
    """Recursively moves a source directory to a target directory, allowing for overwriting existing
        files and directories if enabled.

    Args:
        source: The path to the directory which should be recursively moved.
        target: The path to move the directory to.
        overwrite: Whether to overwrite existing files and directories of the same name, defaults to True.

    Raises:
        FileExistsError: If overwrite is False and a move targets an existing file.
    """
    # Check if paths are the same
    if source == target:
        return

    # Check if target exists
    if target.exists() and overwrite is False:
        raise FileExistsError(f'{target} already exists!')
    if not target.exists():
        mkdir_full_perms(target)

    # Iterate over items in the source directory
    for src, dst in [(source / n, target / n) for n in os.listdir(source)]:

        # Check if destination exists
        if dst.exists() and overwrite is False:
            raise FileExistsError(f'{dst} already exists!')

        # Recurse down next directory or move file
        func = recur_move_dir if src.is_dir() else shutil.move
        func(src, dst)

    # Remove source directory if empty
    if not os.listdir(source):
        os.rmdir(source)


def recur_move_contents(source: Path, target: Path, overwrite: bool = True) -> None:
    """Recursively moves the contents of a source directory to a target directory, allowing for overwriting existing
        files and directories if enabled.

    Args:
        source: The path to the directory which should be recursively moved.
        target: The path to move the directory to.
        overwrite: Whether to overwrite existing files and directories of the same name, defaults to True.

    Raises:
        FileExistsError: If overwrite is False and a move targets an existing file.
    """
    # Check if paths are the same
    if source == target:
        return

    # Ensure target exists
    if not target.exists():
        mkdir_full_perms(target)

    # Move each item to target directory
    for src, dst in [(source / n, target / n) for n in os.listdir(source)]:

        # Handle disallowed existing files
        if dst.exists() and not overwrite:
            raise FileExistsError(f'{dst} already exists!')

        # Move file or directory recursively
        if src.is_dir():
            recur_move_dir(src, dst, overwrite=overwrite)
            continue
        shutil.move(src, dst)


def elevate_contents(path: Path, n: int = 1, overwrite: bool = False) -> Path:
    """Move the contents of a directory up 'n' levels in the directory hierarchy. If 'n' is greater than
        the amount of directories above the current directory, stop at the root.

    Args:
        path: The path to the directory whose contents are to be moved.
        n: The number of levels to move up in the directory hierarchy.
        overwrite: Whether to overwrite existing files with the same name in the target directory, defaults
            to False. If an existing file is encountered while False, raises a FileExistsError.

    Returns:
        Path to the new directory where contents are located.

    Raises:
        FileNotFoundError: If the directory provided doesn't exist.
        ValueError: If "n" was provided as a negative integer.
        FileExistsError: If overwrite is False and a move targets an existing file.
    """
    # Check for bad input
    if not path.is_dir():
        raise FileNotFoundError("The specified directory doesn't exist.")
    if n < 0:
        raise ValueError("The number of levels to raise contents cannot be negative.")

    # Select target, skip if unchanged
    target = path
    for _ in range(n):
        next_dir = target.parent
        if next_dir == target:
            break
        target = next_dir
    if n == 0 or target == path:
        return path

    # Recursively move the contents
    recur_move_contents(path, target, overwrite=overwrite)
    return target


"""
* Context Managers
"""


class DisposableDir:
    def __init__(self, path: Optional[Path] = None):
        """A context manager that creates and returns a temporary directory inside a given path (defaults to current
            directory) and disposes of that directory after the manager closes.

        Args:
            path: Path to create the disposable directory inside.
        """
        path = (path or Path.cwd()) / 'temp'
        name = path.name
        while path.exists():
            path = path.parent / f'{name}_{random.randint(1, 1000000)}'
        self._dir = path

    def __enter__(self) -> Path:
        """Create the disposable directory.

        Returns:
            Path to the disposable directory.
        """
        mkdir_full_perms(self._dir)
        return self._dir

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        """Remove the disposable directory."""
        shutil.rmtree(self._dir)



"""
* Core File Utilities
* Generalized utilities for working with files.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
import hashlib
import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Union

"""
* Generating Files
"""


def ensure_file(path: Path, encoding: str = 'utf-8', boilerplate: str = '') -> None:
    """If a file doesn't exist, create an empty file there.

    Args:
        path: Path to the file.
        encoding: Encoding to use when opening the new file.
        boilerplate: Data to write to the new file.
    """
    if path.is_file():
        return
    with open(path, 'w', encoding=encoding) as f:
        f.write(boilerplate)
    return


def get_temporary_file(path: Path, ext: str = '.tmp', allow_existing: bool = False) -> Path:
    """Check for an existing temporary file representing a provided file path and extension, creates a
        new temporary file if one doesn't exist.

    Args:
        path: Path to the file we plan to download.
        ext: File extension to use when naming the temporary file, defaults to `.tmp`.
        allow_existing: Whether to use existing temporary file if found, defaults to False.

    Returns:
        Path to the temporary file.
    """

    # Look for an existing file
    temp = path.with_suffix(ext)
    if allow_existing:
        for p in os.listdir(path.parent):
            file = path.parent / p
            if file.is_dir() or os.path.samefile(file, path):
                continue
            if temp.name in file.name:
                return file

    # Create a new temporary file
    f = NamedTemporaryFile(prefix=temp.name, dir=temp.parent, delete=False)
    file = f.name
    f.close()
    return Path(file)


"""
* File Naming
"""


def get_unique_filename(path: Path, increment_template: str = '({})') -> Path:
    """If a filepath exists, number the file according to the lowest number that doesn't exist.

    Args:
        path: Path to the file.
        increment_template: String that contains the numeric increment which is increased to arrive at
            a unique filename. Defaults to (i).

    Returns:
        A unique file path.
    """
    stem, i = path.stem, 1
    while path.is_file():
        path = path.with_stem(f'{stem} {increment_template.format(i)}')
        i += 1
    return path


"""
* File Information
"""


def get_file_size_mb(file_path: Union[str, os.PathLike], decimal: int = 1) -> float:
    """Get a file's size in megabytes rounded.

    Args:
        file_path: Path to the file.
        decimal: Number of decimal places to allow when rounding.

    Returns:
        Float representing the filesize in megabytes rounded to decimal length provided.
    """
    return round(os.path.getsize(file_path) / (1024 * 1024), decimal)


def get_sha256(path: Path, chunk_size: int = 4096) -> str:
    """Calculate the SHA-256 hash of a file.

    Args:
        path: Path to the file.
        chunk_size: Max bytes to read from the file on each iteration.

    Returns:
        The SHA-256 hash of this file.
    """
    sha256_hash = hashlib.sha256()
    with open(path, "rb") as f:
        for byte_block in iter(lambda: f.read(chunk_size), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()



"""
* Utils: Images
"""
# Standard Library Imports
import os
from pathlib import Path
from typing import Optional

# Third Party Imports
from PIL import Image
from PIL.Image import Resampling


"""
* Util Funcs
"""


def downscale_image_by_width(
        path_img: Path,
        path_save: Optional[Path] = None,
        max_width: Optional[int] = None,
        optimize: bool = True,
        quality: int | int = 95,
        resample: Resampling = Resampling.LANCZOS,
        convert_rgb: bool = True
) -> Path | None:
    """Downscale an image proportionately to a given maximum width value.

    Args:
        path_img (Path): Path to the image.
        path_save (Path): Path the downscaled image should be saved as. If not provided, will save to:
            /{parent dir}/compressed/{image name}.jpg
        max_width (int): Maximum width of the downscaled image, default: Existing width / 2
        optimize (bool): Whether to use Pillow optimize, default: True
        quality (int): Output image quality, 1-100, default: 95
        resample (Resampling): Resampling algorithm, default: LANCZOS
        convert_rgb (bool): Whether to convert image to RGB, default: True

    Raises:
        OSError: If the image couldn't be downscaled or saved.

    Returns:
        Path to the downscaled image if successful, otherwise None.
    """
    # Open the image, get dimensions
    with Image.open(path_img) as image:

        # Calculate ratio
        if max_width is None:
            max_width = image.width / 2
        ratio = max_width / image.width

    # Forward to ratio-based downscale
    return downscale_image(
        path_img=path_img,
        path_save=path_save,
        ratio=min(ratio, 100),
        optimize=optimize,
        quality=quality,
        resample=resample,
        convert_rgb=convert_rgb)


def downscale_image_by_height(
        path_img: Path,
        path_save: Optional[Path] = None,
        max_height: Optional[int] = None,
        optimize: bool = True,
        quality: int | int = 95,
        resample: Resampling = Resampling.LANCZOS,
        convert_rgb: bool = True
) -> Path | None:
    """Downscale an image proportionately to a given maximum height value.

    Args:
        path_img (Path): Path to the image.
        path_save (Path): Path the downscaled image should be saved as. If not provided, will save to:
            /{parent dir}/compressed/{image name}.jpg
        max_height (int): Maximum height of the downscaled image, default: Existing height / 2
        optimize (bool): Whether to use Pillow optimize, default: True
        quality (int): Output image quality, 1-100, default: 95
        resample (Resampling): Resampling algorithm, default: LANCZOS
        convert_rgb (bool): Whether to convert image to RGB, default: True

    Raises:
        OSError: If the image couldn't be downscaled or saved.

    Returns:
        Path to the downscaled image if successful, otherwise None.
    """
    # Open the image, get dimensions
    with Image.open(path_img) as image:

        # Calculate ratio
        if max_height is None:
            max_height = image.height / 2
        ratio = (max_height / image.height)

    # Forward to ratio-based downscale
    return downscale_image(
        path_img=path_img,
        path_save=path_save,
        ratio=min(ratio, 100),
        optimize=optimize,
        quality=quality,
        resample=resample,
        convert_rgb=convert_rgb)


def downscale_image(
    path_img: Path,
    path_save: Optional[Path] = None,
    ratio: int | float = 50,
    optimize: bool = True,
    quality: int | int = 95,
    resample: Resampling = Resampling.LANCZOS,
    convert_rgb: bool = True
) -> Path | None:
    """Downscale an image proportionately to a given size ratio.

    Args:
        path_img (Path): Path to the image.
        path_save (Path): Path the downscaled image should be saved as. If not provided, will save to:
            /{parent dir}/compressed/{image name}.jpg
        ratio (int | float): Ratio to downscale, default: 50
        optimize (bool): Whether to use Pillow optimize, default: True
        quality (int): Output image quality, 1-100, default: 95
        resample (Resampling): Resampling algorithm, default: LANCZOS
        convert_rgb (bool): Whether to convert image to RGB, default: True

    Raises:
        OSError: If the image couldn't be downscaled or saved.

    Returns:
        Path to the downscaled image if successful, otherwise None.
    """
    # Establish our source and destination directories
    if path_save is None:
        path_save = (path_img.parent / 'compressed' / path_img.name).with_suffix('.jpg')
    if not path_save.parent.is_dir():
        path_save.parent.mkdir(mode=777, parents=True, exist_ok=True)

    # Open the image, get dimensions
    with Image.open(path_img) as image:

        # Convert to RGB
        if convert_rgb:
            image.convert('RGB')

        # Downscale
        if ratio < 100:
            width, height = image.size
            image.thumbnail(
                size=(
                    round(width * ratio),
                    round(height * ratio)),
                resample=resample)

        # Save the new image
        try:
            image.save(
                fp=path_save,
                quality=quality,
                optimize=optimize)
        except (FileExistsError, PermissionError) as e:
            raise OSError("I don't have permission to save the downscaled image.") from e
        except Exception as e:
            raise OSError("Couldn't downscale the provided image.") from e
        return path_save



"""
* Omnitils
* A collection of generic, useful utilities for the Python super-fan on a tight schedule.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
__version__ = '1.4.0'



"""
* Core Fetch Utilities
* General core utilities for making requests.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
import json
from pathlib import Path
from typing import Optional

import requests
from requests import Session

# Default chunk size when using iter_content to save a file
chunk_size_default = 1024 * 1024 * 8

# Default header to pass with requests
request_header_default = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) "
                  "AppleWebKit/537.36 (KHTML, like Gecko) "
                  "Chrome/39.0.2171.95 Safari/537.36"
}


"""
* Session Utility Funcs
"""


def get_new_session(
    headers: Optional[dict] = None,
    stream: bool = False,
    path_cookies: Optional[Path] = None
) -> Session:
    """Returns a Session object equipped with provided features.

    Args:
        headers: Headers object to pass with requests, uses default if not provided.
        stream: Whether to use "stream" when downloading assets.
        path_cookies: Path to cookies file to load saved cookies from, will skip cookies if not provided.

    Returns:
          Session object.
    """
    # Create the Session object
    sess: Session = requests.session()
    if headers is None:
        headers = request_header_default.copy()
    sess.headers = headers
    if stream:
        sess.stream = True

    # Load cookies if provided, return the Session
    if path_cookies and path_cookies.is_file():
        with open(path_cookies, 'r', encoding='utf-8') as f:
            for k, v in json.load(f):
                sess.cookies[k] = v
    return sess



"""
* File Utilities
* Various utilities for working with files, directories, archives, and other file systems.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
from omnitils.files._core import *
from omnitils.files.data import *
from omnitils.files.folders import *



"""
* Omnitils CLI Application
* Internal CLI utilities, primarily used for testing and deployment.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Third Party Imports
import click

# Local Imports
from omnitils._cli.testing import testing_group


@click.group(commands={'test': testing_group})
def omnitils_cli():
    pass


# Export the main CLI Object
OmnitilsCLI = omnitils_cli()
__all__ = ['OmnitilsCLI']



"""
* String Utilities
* Generalized utilities for working with strings.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
# Standard Library Imports
import codecs
from datetime import datetime
import html
import string
from typing import Optional, Union
import unicodedata
from urllib import parse

# Third Party Imports
import yarl
from dateutil import parser

# Maps strings to boolean values
STR_BOOL_MAP = {
    '1': True,
    'y': True,
    't': True,
    'on': True,
    'yes': True,
    'true': True,
    '0': False,
    'n': False,
    'f': False,
    'no': False,
    'off': False,
    'false': False
}

"""
* String Comparisons
"""


def str_to_bool(text: str) -> bool:
    """Converts a truthy string value to a bool. Conversion is case-insensitive.

    Args:
        text: String to check for boolean value.

    Notes:
        - Values are not case-sensitive.
        - True values are 1, t, y, on, yes, and true.
        - False values are 0, f, n, no, off, and true.

    Returns:
        Equivalent boolean value.

    Raises:
        ValueError: If string provided isn't a recognized truthy expression.
    """
    try:
        return STR_BOOL_MAP[text.lower()]
    except KeyError:
        raise ValueError(f"Unrecognized boolean value '{text}'!")


def str_to_bool_safe(text: str, default: bool = False) -> bool:
    """Shorthand for `str_to_bool` which returns default if exception is raised."""
    try:
        return STR_BOOL_MAP[text.lower()]
    except KeyError:
        return default


"""
* String Conversions
"""


def str_to_alnum(text: str, replace_with: str = ' ') -> str:
    """Converts all characters that aren't alphanumeric to spaces or another provided character.

    Args:
        text: String to convert.
        replace_with: Character to replace non-alphanumeric characters with. Defaults to space.

    Return:
        Converted string.
    """
    return ''.join([
        n if n.isalnum() else replace_with
        for n in text
    ])


def normalize_str(text: str, no_space: bool = False) -> str:
    """Normalizes a string for safe comparison.

    Args:
        text: String to normalize.
        no_space: If True remove all spaces, otherwise just leading and trailing spaces.

    Returns:
        Normalized string.
    """
    # Ignore accents and unusual characters
    st = unicodedata.normalize("NFD", text).encode("ascii", "ignore").decode("utf8")

    # Remove spaces?
    st = st.replace(' ', '') if no_space else st.strip()

    # Remove punctuation and make lowercase
    return st.translate(str.maketrans("", "", string.punctuation)).lower()


def normalize_datestr(
        date_str: str,
        date_fmt: str = '%Y-%m-%d',
        date_default: Optional[datetime] = None
) -> str:
    """Normalizes a date string, controlling for potential edge cases.

    Args:
        date_str: Date string to normalize.
        date_fmt: Date format for the output string.
        date_default: Datetime object providing default values for missing
            date and time components in the string

    Returns:
        Normalized date string in the provided format.
    """
    # Replace any invalid integers
    items = [n.strip() for n in str_to_alnum(date_str).split(' ')]
    normalized = ' '.join(['1' if i.isnumeric() and 1 > int(i) else i for i in items])

    # Parse into a date object
    try:
        return parser.parse(
            timestr=normalized,
            default=date_default or datetime(
                year=datetime.today().year,
                month=1,
                day=1,
                hour=0,
                minute=0,
                second=0,
                microsecond=0)
        ).strftime(date_fmt)
    except ValueError as e:
        print(e)
        return datetime.today().strftime(date_fmt)


def normalize_ver(st: str) -> str:
    """Normalize a version string for safe comparison.

    Args:
        st: String to normalize.

    Returns:
        Normalized version string.
    """
    return ''.join([n for n in st if n in '.0123456789'])


"""
* URL Util Funcs
"""


def decode_url(url: str) -> yarl.URL:
    """Unescapes and decodes a URL string and returns it as a URL object.

    Args:
        url: URL string to format.

    Returns:
        Formatted URL object.
    """
    st = codecs.decode(
        html.unescape(parse.unquote(url)),
        'unicode_escape')
    return yarl.URL(st)


"""
* Multiline Utils
"""


def is_multiline(text: Union[str, list[str]]) -> Union[bool, list[bool]]:
    """Check if text or list of texts given contains multiline text (a newline character).

    Args:
        text: String to check or list of strings to check.

    Returns:
        True/False or list of True/False values.
    """
    # String Given
    if isinstance(text, str):
        if '\n' in text or '\r' in text:
            return True
        return False
    # List Given
    if isinstance(text, list):
        return [bool('\n' in t or '\r' in t) for t in text]
    # Invalid data type provided
    raise Exception("Invalid type passed to 'is_multiline', can only accept a string or list of strings.\n"
                    f"Value received: {text}")


def strip_lines(text: str, num: int, sep: str = '\n') -> str:
    """Removes a number of leading or trailing lines from a multiline string.

    Args:
        text: Multiline string.
        num: Positive integer for number leading lines, negative integer for number of trailing lines.
        sep: Separator used to split lines, defaults to '\n'.

    Returns:
        String with lines stripped.
    """
    if num == 0:
        return text
    if num < 0:
        return '\n'.join(text.split(sep)[:num])
    return '\n'.join(text.split(sep)[num:])


def get_line(text: str, i: int, sep: str = '\n') -> str:
    """Get line by index from a multiline string.

    Args:
        text: Multiline string.
        i: Index of the line.
        sep: Separator used to split lines, defaults to '\n'.

    Returns:
        Isolated line.
    """
    if abs(i) > text.count('\n'):
        raise IndexError(f"Not enough lines in multiline string. Index of {i} is invalid.")
    return text.split(sep)[i]


def get_lines(text: str, num: int, sep: str = '\n') -> str:
    """Separate a number of lines from a multiline string.

    Args:
        text: Multiline string.
        num: Number of lines to separate and return, negative integer for trailing lines.
        sep: Newline separator to use for split, defaults to '\n'.

    Returns:
        Isolated lines.
    """
    if num == 0 or abs(num) > text.count('\n') + 1:
        return text
    if num < 0:
        return '\n'.join(text.split(sep)[num:])
    return '\n'.join(text.split(sep)[:num])
